import sys
import re
import math
from transformers import BertTokenizer, BertModel 

import torch.nn as nn
import torch.nn.functional as F
import torch
import numpy as np
import booknlp.common.crf as crf
import booknlp.common.sequence_eval as sequence_eval
from torch.nn import CrossEntropyLoss

class Tagger(nn.Module):

	def __init__(self, freeze_bert=False, base_model=None, tagset=None, supersense_tagset=None, tagset_flat=None, hidden_dim=100, flat_hidden_dim=200, device=None):
		super(Tagger, self).__init__()

		modelName=base_model
		modelName=re.sub("^entities_", "", modelName)
		modelName=re.sub("-v\d.*$", "", modelName)
		matcher=re.search(".*-(\d+)_H-(\d+)_A-.*", modelName)
		bert_dim=0
		modelSize=0
		self.num_layers=0
		if matcher is not None:
			bert_dim=int(matcher.group(2))
			self.num_layers=min(4, int(matcher.group(1)))

			modelSize=self.num_layers*bert_dim

		assert bert_dim != 0
		
		self.tagset=tagset
		self.tagset_flat=tagset_flat

		self.device=device
		self.crf=crf.CRF(len(self.tagset), device)

		self.wn_embedding = nn.Embedding(50, 20)

		self.rev_tagset={tagset[v]:v for v in tagset}
		self.rev_tagset[len(tagset)]="O"
		self.rev_tagset[len(tagset)+1]="O"

		self.num_labels=len(tagset) + 2

		self.supersense_tagset=supersense_tagset
		self.num_supersense_labels=len(supersense_tagset) + 2
		self.supersense_crf=crf.CRF(len(supersense_tagset), device)
		
		self.rev_supersense_tagset={supersense_tagset[v]:v for v in supersense_tagset}
		self.rev_supersense_tagset[len(supersense_tagset)]="O"
		self.rev_supersense_tagset[len(supersense_tagset)+1]="O"

		self.num_labels_flat=len(tagset_flat)

		self.tokenizer = BertTokenizer.from_pretrained(modelName, do_lower_case=False, do_basic_tokenize=False)
		self.bert = BertModel.from_pretrained(modelName)

		self.tokenizer.add_tokens(["[CAP]"], special_tokens=True)
		self.bert.resize_token_embeddings(len(self.tokenizer))

		self.bert.eval()
		
		if freeze_bert:
			for param in self.bert.parameters():
				param.requires_grad = False

		self.hidden_dim = hidden_dim

		self.layered_dropout = nn.Dropout(0.20)

		self.supersense_lstm1 = nn.LSTM(modelSize + 20, hidden_dim, bidirectional=True, batch_first=True)
		self.supersense_hidden2tag1 = nn.Linear(hidden_dim * 2, self.num_supersense_labels)

		self.lstm1 = nn.LSTM(modelSize, hidden_dim, bidirectional=True, batch_first=True)
		self.hidden2tag1 = nn.Linear(hidden_dim * 2, self.num_labels)

		self.lstm2 = nn.LSTM(2*hidden_dim, hidden_dim, bidirectional=True, batch_first=True)
		self.hidden2tag2 = nn.Linear(hidden_dim * 2, self.num_labels)

		self.lstm3 = nn.LSTM(2*hidden_dim, hidden_dim, bidirectional=True, batch_first=True)
		self.hidden2tag3 = nn.Linear(hidden_dim * 2, self.num_labels)
	
		self.flat_dropout = nn.Dropout(0.5)

		self.flat_hidden_dim=flat_hidden_dim
		self.flat_lstm = nn.LSTM(modelSize, self.flat_hidden_dim, bidirectional=True, batch_first=True, num_layers=1)

		self.flat_classifier = nn.Linear(2*self.flat_hidden_dim, self.num_labels_flat)

		param_group = []

		self.bert_params={}
		self.everything_else_params={}

	def forwardFlatSequence(self, input_ids, token_type_ids=None, attention_mask=None, transforms=None, labels=None):

		batch_s, max_len=input_ids.shape

		input_ids = input_ids.to(self.device)
		attention_mask = attention_mask.to(self.device)
		transforms = transforms.to(self.device)

		if labels is not None:
			labels = labels.to(self.device)

		output = self.bert(input_ids, token_type_ids=None, attention_mask=attention_mask, output_hidden_states=True)
		hidden_states=output["hidden_states"]
	
		if self.num_layers == 4:
			all_layers = torch.cat((hidden_states[-1], hidden_states[-2], hidden_states[-3], hidden_states[-4]), 2)
		elif self.num_layers == 2:
			all_layers = torch.cat((hidden_states[-1], hidden_states[-2]), 2)

		out=torch.matmul(transforms,all_layers)

		out, _ = self.flat_lstm(out)

		out=self.flat_dropout(out)
		
		out = out.contiguous().view(-1,out.shape[2])

		logits = self.flat_classifier(out)

		if labels is not None:

			loss_fct = CrossEntropyLoss(ignore_index=-100)
			loss = loss_fct(logits.view(-1, self.num_labels_flat), labels.view(-1))
			return loss

		else:
			return logits

	def forward_supersense(self, wn, input_ids, matrix1, matrix2, attention_mask=None, transforms=None, labels=None, lens=None):
		
		matrix1=matrix1.to(self.device)
		matrix2=matrix2.to(self.device)
		wn=wn.to(self.device)

		wn_embeds=self.wn_embedding(wn)
		
		input_ids = input_ids.to(self.device)
		attention_mask = attention_mask.to(self.device)
		transforms = transforms.to(self.device)

		if lens is not None:
			lens[0] = lens[0].to(self.device)
			lens[1] = lens[1].to(self.device)
			lens[2] = lens[2].to(self.device)

		if labels is not None:
			labels[0] = labels[0].to(self.device)
			labels[1] = labels[1].to(self.device)
			labels[2] = labels[2].to(self.device)
		
		output = self.bert(input_ids, token_type_ids=None, attention_mask=attention_mask, output_hidden_states=True)
		hidden_states=output["hidden_states"]
		if self.num_layers == 4:
			all_layers = torch.cat((hidden_states[-1], hidden_states[-2], hidden_states[-3], hidden_states[-4]), 2)
		elif self.num_layers == 2:
			all_layers = torch.cat((hidden_states[-1], hidden_states[-2]), 2)

		# remove the opening [CLS]
		reduced=torch.matmul(transforms,all_layers)[:,1:,:]

		wn_embeds=wn_embeds[:,1:,:]

		reduced=torch.cat([reduced, wn_embeds], axis=2)

		reduced=self.layered_dropout(reduced)

		lstm_out1, _ = self.supersense_lstm1(reduced)
		tag_space1 = self.supersense_hidden2tag1(lstm_out1)

		to_value=0

		forward_score1 = self.supersense_crf.forward(tag_space1, lens[0]-2)
		sequence_score1 = self.supersense_crf.score(torch.where(labels[0][:,1:] == -100, torch.ones_like(labels[0][:,1:]) * to_value, labels[0][:,1:]), lens[0]-2, logits=tag_space1)
		loss1 = (forward_score1 - sequence_score1).sum()

		return loss1 

	def forward(self, input_ids, matrix1, matrix2, attention_mask=None, transforms=None, labels=None, lens=None):
		
		matrix1=matrix1.to(self.device)
		matrix2=matrix2.to(self.device)
		
		input_ids = input_ids.to(self.device)
		attention_mask = attention_mask.to(self.device)
		transforms = transforms.to(self.device)

		if lens is not None:
			lens[0] = lens[0].to(self.device)
			lens[1] = lens[1].to(self.device)
			lens[2] = lens[2].to(self.device)

		if labels is not None:
			labels[0] = labels[0].to(self.device)
			labels[1] = labels[1].to(self.device)
			labels[2] = labels[2].to(self.device)
		
		output = self.bert(input_ids, token_type_ids=None, attention_mask=attention_mask, output_hidden_states=True)
		hidden_states=output["hidden_states"]
		if self.num_layers == 4:
			all_layers = torch.cat((hidden_states[-1], hidden_states[-2], hidden_states[-3], hidden_states[-4]), 2)
		elif self.num_layers == 2:
			all_layers = torch.cat((hidden_states[-1], hidden_states[-2]), 2)

		# remove the opening [CLS]
		reduced=torch.matmul(transforms,all_layers)[:,1:,:]

		reduced=self.layered_dropout(reduced)

		lstm_out1, _ = self.lstm1(reduced)
		tag_space1 = self.hidden2tag1(lstm_out1)

		input2=torch.matmul(matrix1[:,1:,1:],lstm_out1)

		input2=self.layered_dropout(input2)

		lstm_out2, _ = self.lstm2(input2)
		tag_space2 = self.hidden2tag2(lstm_out2)
		
		input3=torch.matmul(matrix2[:,1:,1:],lstm_out2)

		input3=self.layered_dropout(input3)
		
		lstm_out3, _ = self.lstm3(input3)
		tag_space3 = self.hidden2tag3(lstm_out3)	

		to_value=0

		forward_score1 = self.crf.forward(tag_space1, lens[0]-2)
		sequence_score1 = self.crf.score(torch.where(labels[0][:,1:] == -100, torch.ones_like(labels[0][:,1:]) * to_value, labels[0][:,1:]), lens[0]-2, logits=tag_space1)
		loss1 = (forward_score1 - sequence_score1).sum()

		forward_score2 = self.crf.forward(tag_space2, lens[1]-2)
		sequence_score2 = self.crf.score(torch.where(labels[1][:,1:] == -100, torch.ones_like(labels[1][:,1:]) * to_value, labels[1][:,1:]), lens[1]-2, logits=tag_space2)
		loss2 = (forward_score2 - sequence_score2).sum()

		forward_score3 = self.crf.forward(tag_space3, lens[2]-2)
		sequence_score3 = self.crf.score(torch.where(labels[2][:,1:] == -100, torch.ones_like(labels[2][:,1:]) * to_value, labels[2][:,1:]), lens[2]-2, logits=tag_space3)
		loss3 = (forward_score3 - sequence_score3).sum()

		return loss1 + loss2 + loss3

	def predict_all(self, wn, input_ids, attention_mask=None, transforms=None, lens=None, doEvent=True, doEntities=True, doSS=True):

		def fix(sequence):

			"""
			Ensure tag sequence is BIO-compliant

			"""
			for idx, tag in enumerate(sequence):
				tag=self.rev_tagset[tag]
				if tag.startswith("I-"):
					parts=tag.split("-")
					label=parts[1]
					flag=False
					for i in range(idx-1, -1, -1):
						prev=self.rev_tagset[sequence[i]].split("-")

						if prev[0] == "B" and prev[1] == label:
							flag=True
							break
						
						if prev[0] == "O":
							break

						if prev[0] != "O" and prev[1] != label:
							break

					if flag==False:
						sequence[idx]=self.tagset["B-%s" % label]


		def get_layer_transformation(tag_space, t):

			"""
			After predicting a tag sequence, get the information we need to transform the current layer
			to the next layer (e.g., merging tokens in the same entity and remembering which ones we merged)

			"""

			nl=tag_space.shape[1]

			all_tags=[]
			for tags in t:
				all_tags.append(list(tags.data.cpu().numpy()))

			# matrix for merging tokens in layer n+1 that are part of the same entity in layer n 
			all_index=[]
			# indices of tokens that were merged (so we can restored them later)
			all_missing=[]
			# length of the resulting layer (after merging)
			all_lens=[]

			for tags1 in all_tags:
				fix(tags1)
				index1=self.get_index([tags1], self.rev_tagset)[0]
				for z in range(len(index1)):
					for y in range(len(index1[z]), nl):
						index1[z].append(0)
				for z in range(len(index1), nl):
					index1.append(np.zeros(nl))

				all_index.append(index1)

				missing1=[]
				nll=0
				for idx, tag in enumerate(tags1):
					if idx > 0 and self.rev_tagset[tag].startswith("I-"):
						missing1.append(idx)
					else:
						nll+=1

				all_lens.append(nll)
				all_missing.append(missing1)

			all_index=torch.FloatTensor(np.array(all_index)).to(self.device)

			return all_tags, all_index, all_missing, all_lens

		def supersense_fix(sequence):

			"""
			Ensure tag sequence is BIO-compliant

			"""
			for idx, tag in enumerate(sequence):
				tag=self.rev_supersense_tagset[tag]
				if tag.startswith("I-"):
					parts=tag.split("-")
					label=parts[1]
					flag=False
					for i in range(idx-1, -1, -1):
						prev=self.rev_supersense_tagset[sequence[i]].split("-")

						if prev[0] == "B" and prev[1] == label:
							flag=True
							break
						
						if prev[0] == "O":
							break

						if prev[0] != "O" and prev[1] != label:
							break

					if flag==False:
						sequence[idx]=self.supersense_tagset["B-%s" % label]


		def get_supersense_layer_transformation(tag_space, t):

			"""
			After predicting a tag sequence, get the information we need to transform the current layer
			to the next layer (e.g., merging tokens in the same entity and remembering which ones we merged)

			"""

			nl=tag_space.shape[1]

			all_tags=[]
			for tags in t:
				all_tags.append(list(tags.data.cpu().numpy()))

			# matrix for merging tokens in layer n+1 that are part of the same entity in layer n 
			all_index=[]
			# indices of tokens that were merged (so we can restored them later)
			all_missing=[]
			# length of the resulting layer (after merging)
			all_lens=[]

			for tags1 in all_tags:
				supersense_fix(tags1)
				index1=self.get_index([tags1], self.rev_supersense_tagset)[0]
				for z in range(len(index1)):
					for y in range(len(index1[z]), nl):
						index1[z].append(0)
				for z in range(len(index1), nl):
					index1.append(np.zeros(nl))

				all_index.append(index1)

				missing1=[]
				nll=0
				for idx, tag in enumerate(tags1):
					if idx > 0 and self.rev_supersense_tagset[tag].startswith("I-"):
						missing1.append(idx)
					else:
						nll+=1

				all_lens.append(nll)
				all_missing.append(missing1)

			all_index=torch.FloatTensor(np.array(all_index)).to(self.device)

			return all_tags, all_index, all_missing, all_lens


		all_tags1=all_tags2=all_tags3=event_logits=all_supersense_tags1=None
		
		
		input_ids = input_ids.to(self.device)
		attention_mask = attention_mask.to(self.device)
		transforms = transforms.to(self.device)

		ll=lens.to(self.device)

		sequence_outputs, pooled_outputs, hidden_states = self.bert(input_ids, token_type_ids=None, attention_mask=attention_mask, output_hidden_states=True, return_dict=False)
		if self.num_layers == 4:
			all_layers = torch.cat((hidden_states[-1], hidden_states[-2], hidden_states[-3], hidden_states[-4]), 2)
		elif self.num_layers == 2:
			all_layers = torch.cat((hidden_states[-1], hidden_states[-2]), 2)

		# remove the opening [CLS]
		reduced=torch.matmul(transforms,all_layers)[:,1:,:]

		##
		# ENTITIES
		##

		if doEntities:

			## LAYER 1

			lstm_out1, _ = self.lstm1(reduced)
			tag_space1 = self.hidden2tag1(lstm_out1)

			_, t1 = self.crf.viterbi_decode(tag_space1, ll-2)

			all_tags1, all_index1, all_missing1, n_lens1=get_layer_transformation(tag_space1, t1)

			input2=torch.matmul(all_index1,lstm_out1)

			## LAYER 2

			lstm_out2, _ = self.lstm2(input2)
			tag_space2 = self.hidden2tag2(lstm_out2)
			
			_, t2 = self.crf.viterbi_decode(tag_space2, torch.LongTensor(n_lens1) )

			all_tags2, all_index2, all_missing2, n_lens2=get_layer_transformation(tag_space2, t2)
			
			input3=torch.matmul(all_index2,lstm_out2)

			## LAYER 3

			lstm_out3, _ = self.lstm3(input3)
			tag_space3 = self.hidden2tag3(lstm_out3)

			_, t3 = self.crf.viterbi_decode(tag_space3, torch.LongTensor(n_lens2))

			all_tags3=[]
			for tags in t3:
				all_tags3.append(list(tags.data.cpu().numpy()))

			for tags3 in all_tags3:
				fix(tags3)

			## Insert tokens into later layers that were compressed in earlier layers

			for idx, missing2 in enumerate(all_missing2):
				for m in missing2:
					parts=self.rev_tagset[all_tags3[idx][m-1]].split("-")
					if len(parts) > 1:
						all_tags3[idx].insert(m, self.tagset["I-%s" % parts[1]])
					else:
						all_tags3[idx].insert(m, self.tagset["O"])
			
			for idx, missing1 in enumerate(all_missing1):
				for m in missing1:
					parts=self.rev_tagset[all_tags3[idx][m-1]].split("-")
					if len(parts) > 1:
						all_tags3[idx].insert(m, self.tagset["I-%s" % parts[1]])
					else:
						all_tags3[idx].insert(m, self.tagset["O"])
			
			for idx, missing1 in enumerate(all_missing1):
				for m in missing1:
					parts=self.rev_tagset[all_tags2[idx][m-1]].split("-")
					if len(parts) > 1:
						all_tags2[idx].insert(m, self.tagset["I-%s" % parts[1]])
					else:
						all_tags2[idx].insert(m, self.tagset["O"])

			for idx in range(len(all_tags1)):
				all_tags2[idx]=all_tags2[idx][:len(all_tags1[idx])]
				all_tags3[idx]=all_tags3[idx][:len(all_tags1[idx])]


		###
		# EVENTS
		###

		if doEvent:
			out, _ = self.flat_lstm(reduced)			
			out = out.contiguous().view(-1,out.shape[2])
			event_logits = self.flat_classifier(out)


		##
		# SUPERSENSE
		##

		if doSS:
			wn=wn.to(self.device)
			wn_embeds=self.wn_embedding(wn)
			wn_embeds=wn_embeds[:,1:,:]
			reduced_wn=torch.cat([reduced, wn_embeds], axis=2)

			lstm_out1, _ = self.supersense_lstm1(reduced_wn)
			tag_space1 = self.supersense_hidden2tag1(lstm_out1)

			_, t1 = self.supersense_crf.viterbi_decode(tag_space1, ll-2)

			all_supersense_tags1, all_index1, all_missing1, n_lens1=get_supersense_layer_transformation(tag_space1, t1)


		return all_tags1, all_tags2, all_tags3, event_logits, all_supersense_tags1


	def predict(self, input_ids, attention_mask=None, transforms=None, lens=None):

		def fix(sequence):

			"""
			Ensure tag sequence is BIO-compliant

			"""
			for idx, tag in enumerate(sequence):
				tag=self.rev_tagset[tag]
				if tag.startswith("I-"):
					parts=tag.split("-")
					label=parts[1]
					flag=False
					for i in range(idx-1, -1, -1):
						prev=self.rev_tagset[sequence[i]].split("-")

						if prev[0] == "B" and prev[1] == label:
							flag=True
							break
						
						if prev[0] == "O":
							break

						if prev[0] != "O" and prev[1] != label:
							break

					if flag==False:
						sequence[idx]=self.tagset["B-%s" % label]


		def get_layer_transformation(tag_space, t):

			"""
			After predicting a tag sequence, get the information we need to transform the current layer
			to the next layer (e.g., merging tokens in the same entity and remembering which ones we merged)

			"""

			nl=tag_space.shape[1]

			all_tags=[]
			for tags in t:
				all_tags.append(list(tags.data.cpu().numpy()))

			# matrix for merging tokens in layer n+1 that are part of the same entity in layer n 
			all_index=[]
			# indices of tokens that were merged (so we can restored them later)
			all_missing=[]
			# length of the resulting layer (after merging)
			all_lens=[]

			for tags1 in all_tags:
				fix(tags1)
				index1=self.get_index([tags1], self.rev_tagset)[0]
				for z in range(len(index1)):
					for y in range(len(index1[z]), nl):
						index1[z].append(0)
				for z in range(len(index1), nl):
					index1.append(np.zeros(nl))

				all_index.append(index1)

				missing1=[]
				nll=0
				for idx, tag in enumerate(tags1):
					if idx > 0 and self.rev_tagset[tag].startswith("I-"):
						missing1.append(idx)
					else:
						nll+=1

				all_lens.append(nll)
				all_missing.append(missing1)

			all_index=torch.FloatTensor(all_index).to(self.device)

			return all_tags, all_index, all_missing, all_lens

		## PREDICT

		input_ids = input_ids.to(self.device)
		attention_mask = attention_mask.to(self.device)
		transforms = transforms.to(self.device)

		ll=lens.to(self.device)

		sequence_outputs, pooled_outputs, hidden_states = self.bert(input_ids, token_type_ids=None, attention_mask=attention_mask, output_hidden_states=True, return_dict=False)
		if self.num_layers == 4:
			all_layers = torch.cat((hidden_states[-1], hidden_states[-2], hidden_states[-3], hidden_states[-4]), 2)
		elif self.num_layers == 2:
			all_layers = torch.cat((hidden_states[-1], hidden_states[-2]), 2)

		# remove the opening [CLS]
		reduced=torch.matmul(transforms,all_layers)[:,1:,:]

		## LAYER 1

		lstm_out1, _ = self.lstm1(reduced)
		tag_space1 = self.hidden2tag1(lstm_out1)

		_, t1 = self.crf.viterbi_decode(tag_space1, ll-2)

		all_tags1, all_index1, all_missing1, n_lens1=get_layer_transformation(tag_space1, t1)

		input2=torch.matmul(all_index1,lstm_out1)

		## LAYER 2

		lstm_out2, _ = self.lstm2(input2)
		tag_space2 = self.hidden2tag2(lstm_out2)
		
		_, t2 = self.crf.viterbi_decode(tag_space2, torch.LongTensor(n_lens1) )

		all_tags2, all_index2, all_missing2, n_lens2=get_layer_transformation(tag_space2, t2)
		
		input3=torch.matmul(all_index2,lstm_out2)

		## LAYER 3

		lstm_out3, _ = self.lstm3(input3)
		tag_space3 = self.hidden2tag3(lstm_out3)

		_, t3 = self.crf.viterbi_decode(tag_space3, torch.LongTensor(n_lens2))

		all_tags3=[]
		for tags in t3:
			all_tags3.append(list(tags.data.cpu().numpy()))

		for tags3 in all_tags3:
			fix(tags3)

		## Insert tokens into later layers that were compressed in earlier layers

		for idx, missing2 in enumerate(all_missing2):
			for m in missing2:
				parts=self.rev_tagset[all_tags3[idx][m-1]].split("-")
				if len(parts) > 1:
					all_tags3[idx].insert(m, self.tagset["I-%s" % parts[1]])
				else:
					all_tags3[idx].insert(m, self.tagset["O"])
		
		for idx, missing1 in enumerate(all_missing1):
			for m in missing1:
				parts=self.rev_tagset[all_tags3[idx][m-1]].split("-")
				if len(parts) > 1:
					all_tags3[idx].insert(m, self.tagset["I-%s" % parts[1]])
				else:
					all_tags3[idx].insert(m, self.tagset["O"])
		
		for idx, missing1 in enumerate(all_missing1):
			for m in missing1:
				parts=self.rev_tagset[all_tags2[idx][m-1]].split("-")
				if len(parts) > 1:
					all_tags2[idx].insert(m, self.tagset["I-%s" % parts[1]])
				else:
					all_tags2[idx].insert(m, self.tagset["O"])

		for idx in range(len(all_tags1)):
			all_tags2[idx]=all_tags2[idx][:len(all_tags1[idx])]
			all_tags3[idx]=all_tags3[idx][:len(all_tags1[idx])]

		return all_tags1, all_tags2, all_tags3


	def supersense_predict(self, wn, input_ids, attention_mask=None, transforms=None, lens=None):

		""" Get logits for layered sequence labeling """

		def supersense_fix(sequence):

			"""
			Ensure tag sequence is BIO-compliant

			"""
			for idx, tag in enumerate(sequence):
				tag=self.rev_supersense_tagset[tag]
				if tag.startswith("I-"):
					parts=tag.split("-")
					label=parts[1]
					flag=False
					for i in range(idx-1, -1, -1):
						prev=self.rev_supersense_tagset[sequence[i]].split("-")

						if prev[0] == "B" and prev[1] == label:
							flag=True
							break
						
						if prev[0] == "O":
							break

						if prev[0] != "O" and prev[1] != label:
							break

					if flag==False:
						sequence[idx]=self.supersense_tagset["B-%s" % label]


		def get_layer_transformation(tag_space, t):

			"""
			After predicting a tag sequence, get the information we need to transform the current layer
			to the next layer (e.g., merging tokens in the same entity and remembering which ones we merged)

			"""

			nl=tag_space.shape[1]

			all_tags=[]
			for tags in t:
				all_tags.append(list(tags.data.cpu().numpy()))

			# matrix for merging tokens in layer n+1 that are part of the same entity in layer n 
			all_index=[]
			# indices of tokens that were merged (so we can restored them later)
			all_missing=[]
			# length of the resulting layer (after merging)
			all_lens=[]

			for tags1 in all_tags:
				supersense_fix(tags1)
				index1=self.get_index([tags1], self.rev_supersense_tagset)[0]
				for z in range(len(index1)):
					for y in range(len(index1[z]), nl):
						index1[z].append(0)
				for z in range(len(index1), nl):
					index1.append(np.zeros(nl))

				all_index.append(index1)

				missing1=[]
				nll=0
				for idx, tag in enumerate(tags1):
					if idx > 0 and self.rev_supersense_tagset[tag].startswith("I-"):
						missing1.append(idx)
					else:
						nll+=1

				all_lens.append(nll)
				all_missing.append(missing1)

			all_index=torch.FloatTensor(all_index).to(self.device)

			return all_tags, all_index, all_missing, all_lens

		## PREDICT

		wn=wn.to(self.device)
		input_ids = input_ids.to(self.device)
		attention_mask = attention_mask.to(self.device)
		transforms = transforms.to(self.device)

		ll=lens.to(self.device)

		output = self.bert(input_ids, token_type_ids=None, attention_mask=attention_mask, output_hidden_states=True)
		hidden_states=output["hidden_states"]

		if self.num_layers == 4:
			all_layers = torch.cat((hidden_states[-1], hidden_states[-2], hidden_states[-3], hidden_states[-4]), 2)
		elif self.num_layers == 2:
			all_layers = torch.cat((hidden_states[-1], hidden_states[-2]), 2)

		# remove the opening [CLS]
		reduced=torch.matmul(transforms,all_layers)[:,1:,:]

		wn_embeds=self.wn_embedding(wn)

		wn_embeds=wn_embeds[:,1:,:]

		reduced=torch.cat([reduced, wn_embeds], axis=2)

		lstm_out1, _ = self.supersense_lstm1(reduced)
		tag_space1 = self.supersense_hidden2tag1(lstm_out1)

		_, t1 = self.supersense_crf.viterbi_decode(tag_space1, ll-2)

		all_tags1, all_index1, all_missing1, n_lens1=get_layer_transformation(tag_space1, t1)

		return all_tags1


	def tag_all(self, batched_wn, batched_sents, batched_data, batched_mask, batched_transforms, batched_orig_token_lens, ordering, doEvent=True, doEntities=True, doSS=True):
		
		""" Tag input data for layered sequence labeling """

		c=0
		e=0
		ordered_preds=[]
		ordered_supersense_preds=[]
		ordered_events=[]
		preds_in_order=events_in_order=supersense_preds_in_order=None

		with torch.no_grad():
	
			for b in range(len(batched_data)):
				all_tags1, all_tags2, all_tags3, event_logits, all_supersense_tags1=self.predict_all(batched_wn[b], batched_data[b], attention_mask=batched_mask[b], transforms=batched_transforms[b], lens=batched_orig_token_lens[b], doEvent=doEvent, doEntities=doEntities, doSS=doSS)

				# for each sentence in the batch

				if doEntities:
					for d in range(len(all_tags1)):
			
						preds={}

						for entity in self.get_spans(self.rev_tagset, c, all_tags1[d], batched_orig_token_lens[b][d], batched_sents[b][d][1:]):
							preds[entity]=1
						for entity in self.get_spans(self.rev_tagset, c, all_tags2[d], batched_orig_token_lens[b][d], batched_sents[b][d][1:]):
							preds[entity]=1
						for entity in self.get_spans(self.rev_tagset, c, all_tags3[d], batched_orig_token_lens[b][d], batched_sents[b][d][1:]):
							preds[entity]=1

						ordered_preds.append(preds)

						c+=1

				if doSS:
					for d in range(len(all_supersense_tags1)):

						supersense_preds={}

						for entity in self.get_spans(self.rev_supersense_tagset, e, all_supersense_tags1[d], batched_orig_token_lens[b][d], batched_sents[b][d][1:]):
							supersense_preds[entity]=1
		
						ordered_supersense_preds.append(supersense_preds)
						e+=1

				if doEvent:

					logits=event_logits.cpu()
					ordered_event_preds=[]
					ordered_event_preds += [np.array(r) for r in logits]
					size=batched_wn[b].shape

					logits=logits.view(-1, size[1]-1, 2)

					for row in range(size[0]):
						events={}
						for col in range(batched_orig_token_lens[b][row]-1):
							pred=np.argmax(logits[row][col])
							if pred == 1:
								events[col]=1
						ordered_events.append(events)
				
			if doSS:
				supersense_preds_in_order = [None for i in range(len(ordering))]
				for i, ind in enumerate(ordering):
					supersense_preds_in_order[ind] = ordered_supersense_preds[i]

			if doEntities:
				preds_in_order = [None for i in range(len(ordering))]
				for i, ind in enumerate(ordering):
					preds_in_order[ind] = ordered_preds[i]

			if doEvent:
				events_in_order = [None for i in range(len(ordering))]
				for i, ind in enumerate(ordering):
					events_in_order[ind] = ordered_events[i]


		return preds_in_order, events_in_order, supersense_preds_in_order


	def tag(self, batched_sents, batched_data, batched_mask, batched_transforms, batched_orig_token_lens, ordering):
		
		c=0
		ordered_preds=[]

		with torch.no_grad():
	
			for b in range(len(batched_data)):
				all_tags1, all_tags2, all_tags3=self.predict(batched_data[b], attention_mask=batched_mask[b], transforms=batched_transforms[b], lens=batched_orig_token_lens[b])

				for d in range(len(all_tags1)):
		
					preds={}

					for entity in self.get_spans(self.rev_tagset, c, all_tags1[d], batched_orig_token_lens[b][d], batched_sents[b][d][1:]):
						preds[entity]=1
					for entity in self.get_spans(self.rev_tagset, c, all_tags2[d], batched_orig_token_lens[b][d], batched_sents[b][d][1:]):
						preds[entity]=1
					for entity in self.get_spans(self.rev_tagset, c, all_tags3[d], batched_orig_token_lens[b][d], batched_sents[b][d][1:]):
						preds[entity]=1

					ordered_preds.append(preds)

					c+=1


			preds_in_order = [None for i in range(len(ordering))]
			for i, ind in enumerate(ordering):
				preds_in_order[ind] = ordered_preds[i]

		return preds_in_order

	def supersense_evaluate(self, test_batched_wn, test_batched_sents, test_batched_data, test_batched_mask, test_batched_labels, test_batched_transforms, test_batched_layered_labels1, test_batched_layered_labels2, test_batched_layered_labels3, test_batched_layered_labels4, dev_lens):

		self.eval()

		with torch.no_grad():
			preds={}
			golds={}
			c=0
			for b in range(len(test_batched_data)):
				all_tags1=self.supersense_predict(test_batched_wn[b], test_batched_data[b], attention_mask=test_batched_mask[b], transforms=test_batched_transforms[b], lens=dev_lens[0][b])
				
				for d in range(len(all_tags1)):

					# remove [CLS] from the gold labels
					for entity in self.get_spans(self.rev_supersense_tagset, c, test_batched_layered_labels1[b][d][1:], dev_lens[0][b][d], test_batched_sents[b][d][1:]):
						golds[entity]=1

					# predicted tags already have [CLS] removed
					for entity in self.get_spans(self.rev_supersense_tagset, c, all_tags1[d], dev_lens[0][b][d], test_batched_sents[b][d][1:]):
						preds[entity]=1
	
					c+=1


			F1=sequence_eval.check_span_f1_two_dicts_subcat(golds, preds)

			return F1


	def evaluate(self, test_batched_sents, test_batched_data, test_batched_mask, test_batched_labels, test_batched_transforms, test_batched_layered_labels1, test_batched_layered_labels2, test_batched_layered_labels3, test_batched_layered_labels4, dev_lens):

		""" Evaluate input data (with labels) for layered sequence labeling """

		self.eval()

		with torch.no_grad():
			preds={}
			golds={}
			c=0
			for b in range(len(test_batched_data)):
				all_tags1, all_tags2, all_tags3=self.predict(test_batched_data[b], attention_mask=test_batched_mask[b], transforms=test_batched_transforms[b], lens=dev_lens[0][b])
				
				for d in range(len(all_tags1)):

					# remove [CLS] from the gold labels
					# LitBank has max 4 layers
					for entity in self.get_spans(self.rev_tagset, c, test_batched_layered_labels1[b][d][1:], dev_lens[0][b][d], test_batched_sents[b][d][1:]):
						golds[entity]=1
					for entity in self.get_spans(self.rev_tagset, c, test_batched_layered_labels2[b][d][1:], dev_lens[0][b][d], test_batched_sents[b][d][1:]):
						golds[entity]=1
					for entity in self.get_spans(self.rev_tagset, c, test_batched_layered_labels3[b][d][1:], dev_lens[0][b][d], test_batched_sents[b][d][1:]):
						golds[entity]=1
					for entity in self.get_spans(self.rev_tagset, c, test_batched_layered_labels4[b][d][1:], dev_lens[0][b][d], test_batched_sents[b][d][1:]):
						golds[entity]=1
	
					# predicted tags already have [CLS] removed
					# LitBank has max 4 layers but 4th layer has only 0.1% of tags, so let's just predict 3
					for entity in self.get_spans(self.rev_tagset, c, all_tags1[d], dev_lens[0][b][d], test_batched_sents[b][d][1:]):
						preds[entity]=1
					for entity in self.get_spans(self.rev_tagset, c, all_tags2[d], dev_lens[0][b][d], test_batched_sents[b][d][1:]):
						preds[entity]=1
					for entity in self.get_spans(self.rev_tagset, c, all_tags3[d], dev_lens[0][b][d], test_batched_sents[b][d][1:]):
						preds[entity]=1

					c+=1

			F1=sequence_eval.check_span_f1_two_dicts_subcat(golds, preds)

			return F1

	def evaluateFlat(self, dev_batched_data, dev_batched_mask, dev_batched_labels, dev_batched_transforms, metric, tagset):

		num_labels=len(tagset)

		self.eval()

		with torch.no_grad():

			ordered_preds=[]

			all_preds=[]
			all_golds=[]

			for b in range(len(dev_batched_data)):

				logits = self.forwardFlatSequence(dev_batched_data[b], token_type_ids=None, attention_mask=dev_batched_mask[b], transforms=dev_batched_transforms[b])

				logits=logits.cpu()

				ordered_preds += [np.array(r) for r in logits]
				size=dev_batched_labels[b].shape

				logits=logits.view(-1, size[1], num_labels)

				for row in range(size[0]):
					for col in range(size[1]):
						if dev_batched_labels[b][row][col] != -100:
							pred=np.argmax(logits[row][col])
							all_preds.append(pred.cpu().numpy())
							all_golds.append(dev_batched_labels[b][row][col].cpu().numpy())
			
			return metric(all_golds, all_preds, tagset)


	def tagFlat(self, batched_sents, batched_data, batched_mask, batched_transforms, batched_orig_token_lens, ordering):
		
		with torch.no_grad():

			c=0
			ordered_preds=[]
		
			for b in range(len(batched_data)):

				dataSize=batched_transforms[b].shape
				batch_len=dataSize[0]
				sequence_length=dataSize[1]
				logits = self.forwardFlatSequence(batched_data[b], token_type_ids=None, attention_mask=batched_mask[b], transforms=batched_transforms[b])
				logits=logits.view(-1, sequence_length, self.num_labels_flat)

				logits=logits.cpu()

				preds=np.argmax(logits, axis=2)
				
				for sentence in range(batch_len):
					word_preds=[]

					for idx in range(1,len(batched_sents[b][sentence])-1):
						
						word_preds.append((batched_sents[b][sentence][idx], int(preds[sentence][idx])))

					ordered_preds.append(word_preds)

			preds_in_order = [None for i in range(len(ordering))]
			for i, ind in enumerate(ordering):
				preds_in_order[ind] = ordered_preds[i]

			return preds_in_order


	def get_spans(self, rev_tagset, doc_idx, tags, length, sentence):
		
		# remove the opening [CLS] and closing [SEP]
		tags=tags[:length-2]
		
		entities={}

		for idx, tag in enumerate(tags):

			tag=rev_tagset[int(tag)]

			if tag.startswith("B-"):
				j=idx+1
				parts=tag.split("-")

				while(1):

					if j >= len(tags):
						break

					tagn=rev_tagset[int(tags[j])]
					if tagn.startswith("B") or tagn.startswith("O"):
						break

					parts_n=tagn.split("-")
		
					if parts_n[1] != parts[1]:
						break

					j+=1

				key=doc_idx, parts[1], idx, j

				entities[key]=1

		return entities

	def compress(self, labels, rev_tagset):
		newlabels=[]
		keep=[i for i in range(len(labels[0]))]

		for i in range(len(labels)):
			newlabels.append([labels[i][k] for k in keep])
			newkeep=[]
			for j in keep:
				if labels[i][j] == -100 or not rev_tagset[labels[i][j]].startswith("I-"):
					newkeep.append(j)

			keep=newkeep
		
		return newlabels

	def get_index(self, all_labels, rev_tagset):
		indices=[]
		for labels in all_labels:
			index=[]
			n=len(labels)
			for idx, label in enumerate(labels):
				ind=list(np.zeros(n))

				if label == -100 or not rev_tagset[label].startswith("I-"):
					ind[idx]=1		
					index.append(ind)
				else:
					index[-1][idx]=1
		
			indices.append(index)

		for index in indices:
			for i, idx in enumerate(index):
				idx=idx/np.sum(idx)

				index[i]=list(idx)

		return indices




import numpy as np
import torch

def get_batches(model, sentences, max_batch, tagset, training=True):

	"""
	Partitions a list of sentences (each a list containing [word, label]) into a set of batches
	Returns:

	-- batched_sents: original tokens in sentences
		
	-- batched_orig_token_lens: length of original tokens in sentences

	-- batched_data: token ids of sentences. [[101 37 42 102], [101 7 102 0]]
	
	-- batched_mask: Binary flag for real tokens (1) and padded tokens (0) [[1 1 1 1], [1 1 1 0]] (for BERT)
	
	-- batched_transforms: BERT word piece tokenization splits words into pieces; this matrix specifies how
	to combine those pieces back into the original tokens (by averaging their representations) using matrix operations.
	If the original sentence is 3 words that have been tokenized into 4 word piece tokens [101 37 42 102] 
	(where 37 42 are the pieces of one original word), the transformation matrix is 4 x 3 (zero padded to 4 x 4), 
	resulting in the original sequence length of 3. [[1 0 0 0], [0 0.5 0.5 0], [0 0 0 1]]. 

	-- batched_labels: Labels for each sentence, one label per original token (prior to word piece tokenization). Padded tokens
		and [CLS] and [SEP] have labels -100.

	-- batched_layered_labels{1,2,3,4,5}: Labels for each sentence, one label per original token (prior to word piece tokenization). Padded tokens and [CLS] and [SEP] have labels -100.

	-- batched_index{1,2,3}: For nested NER, words that are part of the same entity in layer n are merged together in layer n+1. batched_index is a matrix that specifies how to combine those token representations when moving between layers.  The matrix is 0-padded to be square in the length of the input layer.
	
	-- batched_newlabel{1,2,3}.  batched_labels (above) specifies the labels for each absolute token position; for nested NER, however, the sequence length gets smaller when moving from lower layers to higher layers (since tokens within the same entity are combined).  batched_newlabel specifies the correct labels for the actual sequence length in a given layer.  Padded to the max sequence length for a batch with -100.

	-- batched_lens{1,2,3}: The actual sequence length in a given layer.

	-- ordering: inverse argsort to recover original ordering of sentences.

	"""

	rev_tagset={tagset[v]:v for v in tagset}

	maxLen=0

	all_sents=[]
	all_orig_token_lens=[]

	for sentence in sentences:
		length=0
		# ts=[x[0] for x in sentence]
		ts=[' '.join(x) for x in sentence]
		all_sents.append(ts)

		all_orig_token_lens.append(len(sentence))

		for toks in sentence:
			# toks=model.tokenizer.tokenize(word[0])
			length+=len(toks)

		if length > maxLen:
			maxLen=length

	all_data=[]
	all_masks=[]
	all_labels=[]


	all_layered_labels1=[]
	all_layered_labels2=[]
	all_layered_labels3=[]
	all_layered_labels4=[]
	all_layered_labels5=[]
	
	all_indices=[]
	all_newlabels=[]

	all_transforms=[]

	for sentence in sentences:
		tok_ids=[]
		input_mask=[]
		labels=[]

		layered_labels1=[]
		layered_labels2=[]
		layered_labels3=[]
		layered_labels4=[]
		layered_labels5=[]
		
		transform=[]

		all_toks=[]
		n=0
		for idx, toks in enumerate(sentence):
			# toks=model.tokenizer.tokenize(word[0])
			all_toks.append(toks)
			n+=len(toks)

		cur=0
		for idx, word in enumerate(sentence):
			toks=all_toks[idx]
			ind=list(np.zeros(n))
			for j in range(cur,cur+len(toks)):
				ind[j]=1./len(toks)
			cur+=len(toks)
			transform.append(ind)

			tok_ids.extend(model.tokenizer.convert_tokens_to_ids(toks))

			input_mask.extend(np.ones(len(toks)))

			if training:

				labels.append(int(word[1]))
				layered_labels1.append(int(word[1]))
				layered_labels2.append(int(word[2]))
				layered_labels3.append(int(word[3]))
				layered_labels4.append(int(word[4]))
				layered_labels5.append(int(word[5]))

		all_data.append(tok_ids)
		all_masks.append(input_mask)
		all_transforms.append(transform)

		if training:

			all_labels.append(labels)
			all_layered_labels1.append(layered_labels1)
			all_layered_labels2.append(layered_labels2)
			all_layered_labels3.append(layered_labels3)
			all_layered_labels4.append(layered_labels4)
			all_layered_labels5.append(layered_labels5)

			newlabels=model.compress([layered_labels1, layered_labels2, layered_labels3])
			indices=model.get_index(newlabels)
			
			all_indices.append(indices)
			all_newlabels.append(newlabels)


	lengths = np.array([len(l) for l in all_data])
	ordering = np.argsort(lengths)
	ordered_data = [None for i in range(len(all_data))]
	ordered_masks = [None for i in range(len(all_data))]
	ordered_transforms = [None for i in range(len(all_data))]
	orig_sents = [None for i in range(len(all_data))]
	orig_token_lens = [None for i in range(len(all_data))]

	if training:

		ordered_labels = [None for i in range(len(all_data))]
		ordered_layered_labels1 = [None for i in range(len(all_data))]
		ordered_layered_labels2 = [None for i in range(len(all_data))]
		ordered_layered_labels3 = [None for i in range(len(all_data))]
		ordered_layered_labels4 = [None for i in range(len(all_data))]
		ordered_layered_labels5 = [None for i in range(len(all_data))]

		ordered_indices = [None for i in range(len(all_data))]
		ordered_newlabels = [None for i in range(len(all_data))]

	for i, ind in enumerate(ordering):
		ordered_data[i] = all_data[ind]
		ordered_masks[i] = all_masks[ind]
		orig_sents[i]=all_sents[ind]
		orig_token_lens[i]=all_orig_token_lens[ind]
		ordered_transforms[i] = all_transforms[ind]

		if training:

			ordered_labels[i] = all_labels[ind]
			ordered_layered_labels1[i] = all_layered_labels1[ind]
			ordered_layered_labels2[i] = all_layered_labels2[ind]
			ordered_layered_labels3[i] = all_layered_labels3[ind]
			ordered_layered_labels4[i] = all_layered_labels4[ind]
			ordered_layered_labels5[i] = all_layered_labels5[ind]

			ordered_indices[i] = all_indices[ind]
			ordered_newlabels[i] = all_newlabels[ind]

	batched_data=[]
	batched_mask=[]
	batched_labels=[]
	batched_transforms=[]
	batched_indices=[]
	batched_layered_labels1=[]
	batched_layered_labels2=[]
	batched_layered_labels3=[]
	batched_layered_labels4=[]
	batched_layered_labels5=[]
	
	batched_index1=[]
	batched_index2=[]
	batched_index3=[]

	batched_newlabel1=[]
	batched_newlabel2=[]
	batched_newlabel3=[]

	batched_sents=[]
	batched_orig_token_lens=[]

	batched_lens1=[]
	batched_lens2=[]
	batched_lens3=[]

	i=0

	current_batch=max_batch
	order_to_batch_map=[]
	batch_num=0

	while i < len(ordered_data):

		for j in range(current_batch):
			order_to_batch_map.append((batch_num, current_batch, j))

		batch_num+=1

		batch_data=ordered_data[i:i+current_batch]
		batch_mask=ordered_masks[i:i+current_batch]
		batch_sents=orig_sents[i:i+current_batch]
		batch_orig_lens=orig_token_lens[i:i+current_batch]
		batch_transforms=ordered_transforms[i:i+current_batch]
		
		max_len = max([len(sent) for sent in batch_data])
		max_label_length = max([l for l in batch_orig_lens])

		if training:

			batch_labels=ordered_labels[i:i+current_batch]
			
			batch_layered_labels1=ordered_layered_labels1[i:i+current_batch]
			batch_layered_labels2=ordered_layered_labels2[i:i+current_batch]
			batch_layered_labels3=ordered_layered_labels3[i:i+current_batch]
			batch_layered_labels4=ordered_layered_labels4[i:i+current_batch]
			batch_layered_labels5=ordered_layered_labels5[i:i+current_batch]

			batch_indices=ordered_indices[i:i+current_batch]
			batch_newlabels=ordered_newlabels[i:i+current_batch]


		batch_index1=[]
		batch_index2=[]
		batch_index3=[]
		
		batch_new_label1=[]
		batch_new_label2=[]
		batch_new_label3=[]

		lens1=[]
		lens2=[]
		lens3=[]

		for j in range(len(batch_data)):
			
			blen=len(batch_data[j])

			for k in range(blen, max_len):
				batch_data[j].append(0)
				batch_mask[j].append(0)
				for z in range(len(batch_transforms[j])):
					batch_transforms[j][z].append(0)

			for k in range(len(batch_transforms[j]), max_label_length):
				batch_transforms[j].append(np.zeros(max_len))

			if training:

				blab=len(batch_labels[j])
				indexes=batch_indices[j]				
				newlabelz=batch_newlabels[j]

				lens1.append(len(indexes[0][0]))
				lens2.append(len(indexes[1][0]))
				lens3.append(len(indexes[2][0]))

				for k in range(3):
					indexk=indexes[k]

					# pad each row with zeros to the same length
					for y in range(len(indexk)):
						for z in range(len(indexk[y]), max_label_length):
							indexk[y].append(0)

					# pad the matrix with zeros to be square
					for y in range(len(indexk), max_label_length):
						indexk.append(np.zeros(max_label_length))

					# pad the reduced labels with -100
					for z in range(len(newlabelz[k]), max_label_length):
						newlabelz[k].append(-100)

				batch_index1.append(indexes[0])
				batch_index2.append(indexes[1])
				batch_index3.append(indexes[2])

				batch_new_label1.append(newlabelz[0])
				batch_new_label2.append(newlabelz[1])
				batch_new_label3.append(newlabelz[2])


				for k in range(blab, max_label_length):
					batch_labels[j].append(-100)

					batch_layered_labels1[j].append(-100)
					batch_layered_labels2[j].append(-100)
					batch_layered_labels3[j].append(-100)
					batch_layered_labels4[j].append(-100)
					batch_layered_labels5[j].append(-100)



		batched_data.append(torch.LongTensor(batch_data))
		batched_mask.append(torch.FloatTensor(batch_mask))
		batched_sents.append(batch_sents)
		batched_orig_token_lens.append(torch.LongTensor(batch_orig_lens))

		batched_transforms.append(torch.FloatTensor(np.array(batch_transforms)))

		if training:

			batched_labels.append(torch.LongTensor(batch_labels))

			batched_layered_labels1.append(torch.LongTensor(batch_layered_labels1))
			batched_layered_labels2.append(torch.LongTensor(batch_layered_labels2))
			batched_layered_labels3.append(torch.LongTensor(batch_layered_labels3))
			batched_layered_labels4.append(torch.LongTensor(batch_layered_labels4))
			batched_layered_labels5.append(torch.LongTensor(batch_layered_labels5))

			batched_index1.append(torch.FloatTensor(batch_index1))
			batched_index2.append(torch.FloatTensor(batch_index2))
			batched_index3.append(torch.FloatTensor(batch_index3))


			batched_newlabel1.append(torch.LongTensor(batch_new_label1))
			batched_newlabel2.append(torch.LongTensor(batch_new_label2))
			batched_newlabel3.append(torch.LongTensor(batch_new_label3))

		batched_lens1.append(torch.LongTensor(lens1))
		batched_lens2.append(torch.LongTensor(lens2))
		batched_lens3.append(torch.LongTensor(lens3))
		
		i+=current_batch

		if max_len > 100:
			current_batch=12
		if max_len > 200:
			current_batch=6
	
	
	if training:
		return batched_sents, batched_data, batched_mask, batched_labels, batched_transforms, ordering, batched_layered_labels1, batched_layered_labels2, batched_layered_labels3, batched_layered_labels4, batched_layered_labels5, batched_index1, batched_index2, batched_index3, batched_newlabel1, batched_newlabel2, batched_newlabel3, [batched_lens1, batched_lens2, batched_lens3]

		
	return batched_sents, batched_data, batched_mask, batched_transforms, batched_orig_token_lens, ordering, order_to_batch_map





"""
Code for inferring the referential gender of name-clustered characters (e.g., ELIZABETH BENNETT = {Elizabeth, Lizzy, Elizabeth Bennett}) and common NPs (e.g "the boy") solely from the gendered pronouns (he/his/him/she/her/they/them/ze/xe/etc.) that appear near them. This method uses the IBM model 1 EM alignment algorithm (originally used for MT) to align entities with pronoun tokens, starting from a prior that all entities are gender-neutral and then updating that belief based on their explanatory power for the pronouns that surround them.

This method encodes several assumptions:

* This method describes the referential gender of characters, and not their gender identity. Characters are described by the pronouns used to refer to them (e.g., he/him, she/her) rather than labels like "M/F".

* Prior information on the alignment of names with referential gender (e.g., from government records or larger background datasets) can be used to provide some information to inform this process if desired (e.g., "Tom" is often associated with he/him in pre-1923 English texts), but should be updateable *in the context of a specific book* (e.g., "Tom" in the book "Tom and Some Other Girls", where Tom is aligned with she/her). 

* Users should be free to define the referential gender categories used here.  The default is {he, him, his}, 
{she, her} and {they, them, their}, but can be expanded to include categories such as {xe, xem, xyr, xir}, {ze, zem, zir, hir}, etc.


"""


# requires scipy==1.5.4
from collections import Counter
import sys
from tqdm import tqdm
from os import listdir
from os.path import isfile
import os
from booknlp.common.pipelines import Token

import random

class GenderEM:

	def __init__(self, outfile=None, tokens=None, entities=None, entityFiles=None, tokenFiles=None, hyperparameterFile=None, distance=25, num_epochs=25, refs=None, upper=10, use_tagged_pronouns_only=True, genders=[["he", "him", "his"],["she", "her"],	["they", "them", "their"]] ):

		# Number of epochs for EM
		self.num_epochs=num_epochs

		# Candidates entities must within this number of preceding tokens of pronouns
		self.distance=distance

		# maximum pseudocount value (set lower to contrain how much influence priors have)
		self.upper=upper

		self.honorifics={"mr.":1, "mrs.":1, "miss":1, "lady":1, "sir":1, "captain":1, "mr":1, "lord":1, "aunt":1, "madame":1, "mrs":1, "uncle":1, "colonel":1, "monsieur":1, "mademoiselle":1, "general":1, "major":1, "sergeant":1, "ms.":1,  "king":1, "queen":1, "herr":1, "frau":1, "fräulein":1, "dame":1, "mister":1, "master":1, "mistress":1, "prince":1, "princess":1, "lieutenant":1 }
		self.honorific_priors={}
		self.use_tagged_pronouns_only=use_tagged_pronouns_only
		self.genders=genders

		self.genderID={}
		self.gender_pronouns={}

		self.num_genders=len(self.genders)

		for idx, gender in enumerate(self.genders):
			self.genderID["/".join(gender)]=idx
			for term in gender:
				self.gender_pronouns[term]=idx

		self.reverseID={self.genderID[k]:k for k in self.genderID}
		

		# f = he/she/they
		# e = John, Kate, the man, her husband, his mother

		self.joint_e_f_counts={}
		self.e_counts={}

		self.t_f_e={}

		self.hyperparameters={}

		if hyperparameterFile is not None:
			self.read_hyperparams(hyperparameterFile)

		self.vocab={}

		if entityFiles is not None and tokenFiles is not None: 
			self.build_vocab_from_files(entityFiles, tokenFiles)
		elif tokens is not None and entities is not None and refs is not None:
			self.build_vocab(tokens, entities, refs)


		self.add_hyperparameters_to_counts(refs=refs, entities=entities, tokens=tokens)

		self.maximization()
		self.delete_counts()

		self.outfile=outfile


	def add_hyperparameters_to_counts(self, refs=None, entities=None, tokens=None):
		for e in self.vocab:

			mf=[1]*self.num_genders

			# add pseudocounts if we have them from the hyperparameters
			if self.hyperparameters is not None and e in self.hyperparameters:
				mf=self.hyperparameters[e]

			for f in range(self.num_genders):
				self.joint_e_f_counts[e,f]=mf[f] + 0.1
				if e not in self.e_counts:
					self.e_counts[e]=0
				self.e_counts[e]+=mf[f] + 0.1
		

		# for entities/coref IDs, update the coref ID to include priors on the names associated with that ID
		# e.g. 17 -> Jane Bennett, Jane, etc. -- we want to include our priors on "Jane" within entity 17

		counts={}
		if refs is not None:
			for idx, ref in enumerate(refs):
				start, end, cat, text=entities[idx]

				prop=cat.split("_")[0]
				cat=cat.split("_")[1]

				phraseHead=None

				if prop == "NOM":
					phraseHead=self.get_head(start, end, tokens)

				if phraseHead is not None:
					key="%s\t%s" % (tokens[phraseHead].text, prop)
				else:
					key="%s\t%s" % (text, prop)

				if ref != -1:
					if ref not in counts:
						counts[ref]=Counter()
					counts[ref][key.lower()]+=1

		for e in self.vocab:
			mf=[0.]*self.num_genders
			parts=e.split("\t")
			if parts[1].lower() == "coref":
				idd=int(parts[0])

				if idd != -1:

					for text in counts[idd]:

						# add prior over entire name if present (e.g., "[Tom]")
						for f in range(self.num_genders):
							if (text,f) in self.joint_e_f_counts:
								mf[f]=self.joint_e_f_counts[text,f] * counts[idd][text]

					if sum(mf) > self.upper:
						for i in range(self.num_genders):
							mf[i]=(mf[i]/sum(mf)) * self.upper

					for text in counts[idd]:

						# add prior over honorifics if present (e.g., "[mr.] sawyer")
						first_name_token=text.split(" ")[0]
						if first_name_token in self.honorific_priors:
							for f in range(self.num_genders):
								mf[f]=self.honorific_priors[first_name_token][f] * counts[idd][text] * 10000


					self.e_counts[e]=0
					for f in range(self.num_genders):
						self.joint_e_f_counts[e,f]=mf[f] + 0.1
						self.e_counts[e]+=mf[f] + 0.1

	def read_hyperparams(self, filename):
		self.hyperparameters={}
		with open(filename) as file:
			header=file.readline().rstrip()
			gender_mapping={}
			for idx, val in enumerate(header.split("\t")[2:]):
				if val in self.genderID:
					gender_mapping[self.genderID[val]]=idx+2

				else:
					print("NOTE PRIOR CATEGORY %s NOT AMONG GENDERS" % val)

			for line in file:
				cols=line.rstrip().split("\t")
				term=cols[0]
				proper=cols[1]

				vals=[0]*(self.num_genders)
				for val in gender_mapping:
					vals[val]=float(cols[gender_mapping[val]])

				first_token=term.split(" ")[0]
				if first_token in self.honorifics:
					if first_token not in self.honorific_priors:
						self.honorific_priors[first_token]=[0]*self.num_genders

					for i in range(self.num_genders):
						self.honorific_priors[first_token][i]+=vals[i]

				total=sum(vals)
				if total >= self.upper:
					for i in range(len(vals)):
						vals[i]=(vals[i]/total) * self.upper
					
					self.hyperparameters[("%s\t%s" % (term, proper)).lower()]=vals

			for honorific in self.honorific_priors:
				total=sum(self.honorific_priors[honorific])
				if total >= self.upper:
					for i in range(self.num_genders):
						self.honorific_priors[honorific][i]=(self.honorific_priors[honorific][i]/total) * self.upper
					
	def get_head(self, start, end, tokens):
		phraseHead=None
		for idd in range(start, end+1):
			head=tokens[idd].dephead
			if head < start or head > end:
				phraseHead=idd
		return phraseHead

	def build_vocab(self, tokens, entities, refs=None):


		for term in self.hyperparameters:
			self.vocab[term]=1

		for idx, (start, end, cat, text) in enumerate(entities):

			prop=cat.split("_")[0]
			cat=cat.split("_")[1]

			if cat == "PER":

				if refs is not None and refs[idx] != -1:
					key="%s\t%s" % (refs[idx], "COREF")

				else:

					phraseHead=None

					if prop == "NOM":
						phraseHead=self.get_head(start, end, tokens)

					if phraseHead is not None:
						key="%s\t%s" % (tokens[phraseHead].text, prop)
					else:
						key="%s\t%s" % (text, prop)


				self.vocab[key.lower()]=1					


	def build_vocab_from_files(self, entityFiles, tokenFiles):
	

		for term in self.hyperparameters:
			self.vocab[term]=1

		for idx in tqdm(range(len(entityFiles))):

			entities=self.read_entities(entityFiles[idx])
			tokens=self.read_tokens(tokenFiles[idx])

			self.build_vocab(tokens, entities)



	def tagFromFile(self, entityFiles, tokenFiles):
		
		all_X=[]
		all_Y=[]

		for idx in tqdm(range(len(entityFiles))):

			entities=self.read_entities(entityFiles[idx])
			tokens=self.read_tokens(tokenFiles[idx])

			X, Y=self.process(tokens, entities)
			all_X.append(X)
			all_Y.append(Y)


		for epoch in range(self.num_epochs):
			
			for i in tqdm(range(len(all_X))):
				for e, f in zip(all_X[i], all_Y[i]):
					self.update(e,f)

			self.maximization()
			self.print(epoch)
			self.delete_counts()
			self.add_hyperparameters_to_counts()


	def tag(self, entities, tokens, refs, minThreshold=0):

		X, Y=self.process(tokens, entities, refs)	

		for epoch in range(self.num_epochs):
			
			for e, f in zip(X, Y):
				self.update(e,f)

			self.maximization()
			if epoch < self.num_epochs-1:
				self.delete_counts()
				self.add_hyperparameters_to_counts(refs=refs, entities=entities, tokens=tokens)


		genders={}
		for e, f in self.t_f_e:
			if f == 1:

				vals={}
				total=0
				for i in range(self.num_genders):
					vals[self.reverseID[i]] = self.joint_e_f_counts[e,i]
					total+=vals[self.reverseID[i]]

				if total > 0:
					for val in vals:
						vals[val]=float("%.3f" % (vals[val]/total))

				maxID=None
				maxVal=0
				for val in vals:
					if vals[val] > maxVal:
						maxVal=vals[val]
						maxID=val

				cat=e.split("\t")[1]
				if cat == "coref":
					idd=int(e.split("\t")[0])
					genders[idd]={"inference":vals, "argmax":maxID, "max":maxVal, "total":float("%.3f" % total)}

		return genders

	def print(self, epoch):
		with open("%s.%s" % (self.outfile, epoch), "w", encoding="utf-8") as out:
			valstr=[]
			for gender in self.genders:
				valstr.append('/'.join(gender))

			out.write("%s\t%s\t%s\n" % ("term", "proper", '\t'.join(valstr)))

			for e, f in self.t_f_e:
				if f == 1:
					vals=[0]*self.num_genders
					for i in range(self.num_genders):
						vals[i] = self.joint_e_f_counts[e,i]

					out.write("%s\t%s\n" % (e, '\t'.join(["%.3f" % x for x in vals])))


	# get all entities in a window *before* a target pronoun
	def get_mentions(self, loc_starts, idx):

		mentions=[]

		for i in range(1,self.distance):
			j=idx-i
			if j in loc_starts:
				for end, comp, text in loc_starts[j]:
					# skip entities that enclose the pronoun 
					if end < idx:
						mentions.append((j, end, comp, text))

		return mentions


	def update(self, e_seq, f_seq):

		for i, f in enumerate(f_seq):
			total=0
			for j, e in enumerate(e_seq):
				total+=self.t_f_e[e, f]
			for j, e in enumerate(e_seq):
				delta_k_i_j=self.t_f_e[e, f] / total

				self.joint_e_f_counts[e,f]+=delta_k_i_j
				self.e_counts[e]+=delta_k_i_j


	def maximization(self, delete_counts=True):

		for (e,f) in self.joint_e_f_counts:

			if self.e_counts[e] > 0:
				self.t_f_e[e,f]=self.joint_e_f_counts[e,f]/self.e_counts[e]
			else:
				self.t_f_e[e,f]=0

			parts=e.split("\t")

	def delete_counts(self):
		for (e,f) in self.joint_e_f_counts:
			self.joint_e_f_counts[e,f]=0
		
		for e in self.e_counts:
			self.e_counts[e]=0


	def process(self, toks, entities, refs=None):

		# organize the token start positions for each entity so we can find them faster
		loc_starts={}

		tagged_pronouns=[]

		for idx, (start, end, cat, text) in enumerate(entities):
			prop=cat.split("_")[0]
			ner_type=cat.split("_")[1]

			if ner_type == "PER":

				if text.lower() in self.gender_pronouns:
					tagged_pronouns.append(start)

				# if this mention is already coreferent (e.g., through name clustering), use that coref ID instead
				if refs is not None and refs[idx] != -1:
					key="%s\t%s" % (refs[idx], "COREF")

				else:
					phraseHead=None

					if prop == "NOM":
						phraseHead=self.get_head(start, end, toks)

					if phraseHead is not None:
						key="%s\t%s" % (toks[phraseHead].text, prop)
					else:
						key="%s\t%s" % (text, prop)


				if start not in loc_starts:
					loc_starts[start]=[]

				loc_starts[start].append((end, idx, key))

		X=[]
		Y=[]

		# for each gendered pronoun in the text, identify all entity/common NP mentions in a window around it
		if self.use_tagged_pronouns_only:
			for idx in tagged_pronouns:
				tok=toks[idx].text
				mentions=self.get_mentions(loc_starts, idx)
				mention_refs=[]
				for start, end, loc_idx, text in mentions:
					mention_refs.append(text.lower())

				if len(mention_refs) > 0:
					X.append(mention_refs)
					gender=self.gender_pronouns[tok.lower()]
					Y.append([gender])

		else:
			for idx, tokenObject in enumerate(toks):
				tok=tokenObject.text
				if tok.lower() in self.gender_pronouns:

					mentions=self.get_mentions(loc_starts, idx)
					mention_refs=[]
					for start, end, loc_idx, text in mentions:
						mention_refs.append(text.lower())

					if len(mention_refs) > 0:
						X.append(mention_refs)
						gender=self.gender_pronouns[tok.lower()]
						Y.append([gender])


		return X, Y




	def read_tokens(self, tokenFile):
		toks=[]
		with open(tokenFile) as file:
			for line in file:
				cols=line.rstrip().split("\t")

				paragraph_id=int(cols[0])
				sentence_id=int(cols[1])
				index_within_sentence_idx=int(cols[2])
				token_id=int(cols[3])
				text=cols[4]
				lemma=cols[5]
				startByte=int(cols[6])
				pos=cols[8]
				deprel=cols[9]
				dephead=int(cols[10])
				ner=None

				tok=Token(paragraph_id, sentence_id, index_within_sentence_idx, token_id, text, pos, None, lemma, deprel, dephead, ner, startByte)

				toks.append(tok)

		return toks


	def read_entities(self, entityFile):

		entities=[]

		with open(entityFile) as file:

			for line in file:
				cols=line.rstrip().split("\t")
				start=int(cols[0])
				end=int(cols[1])
				prop=cols[2]
				cat=cols[3]
				cat="%s_%s" % (prop, cat)
				text=cols[4]

				entities.append((start, end, cat, text))

		return entities


	def update_gender_from_coref(self, genders, entities, corefs):

		counts={}

		for idx, (start, end, cat, text) in enumerate(entities):
			coref=corefs[idx]
			if coref not in counts:
				counts[coref]={}

			for gc in self.genders:
				for term in gc:
					if text.lower() == term:
						key='/'.join(gc)
						if key not in counts[coref]:
							counts[coref][key]=0

						counts[coref][key]+=1

		for c in counts:
			if c not in genders:
			

				total=0
				maxg=None
				maxv=0
				for key in counts[c]:
					total+=counts[c][key]
					if counts[c][key] > maxv:
						maxv=counts[c][key]
						maxg=key

				if total == 0:
					continue

				genders[c]={}
				genders[c]["inference"]={}
				for i in range(self.num_genders):
					genders[c]["inference"][self.reverseID[i]] = 0

				for key in counts[c]:
					genders[c]["inference"][key]=counts[c][key]/total

				if total >0:
					genders[c]["argmax"]=maxg
					genders[c]["max"]=maxv/total
					genders[c]["total"]=total


		return genders

if __name__ == "__main__":

	top=sys.argv[1]
	outfile=sys.argv[2]
	hyperparams=None
	if len(sys.argv) > 3:
		hyperparams=sys.argv[3]
	onlyfiles = [f for f in listdir(top)]
	num=len(onlyfiles)

	ent_files=[]
	tok_files=[]
	for idd in (onlyfiles[:num]):
		entitiyFile=os.path.join(top, idd, "%s.entities" % idd)
		tokensFile=os.path.join(top, idd, "%s.tokens" % idd)
		if isfile(entityFile) and isfile(tokensFile):
			ent_files.append(entityFile)
			tok_files.append(tokensFile)

	genders=[ ["he", "him", "his"], ["she", "her"], ["they", "them", "their"] ] 
	genderEM=GenderEM(outfile=outfile, entityFiles=ent_files, tokenFiles=tok_files, hyperparameterFile=hyperparams, genders=genders)
	genderEM.tagFromFile(ent_files, tok_files)




from setuptools import setup, find_packages

setup(name='booknlp', 
	version='1.0.7.1', 
	packages=find_packages(),
	py_modules=['booknlp'],
	url="https://github.com/dbamman/book-nlp",
	author="David Bamman",
	author_email="dbamman@berkeley.edu",
	include_package_data=True, 
	license="MIT",
	install_requires=['torch>=1.7.1',
					  #'tensorflow>=1.15',
					  'spacy>=3',
                      'transformers>=4.11.3,<=4.30.0'         
                      ],

	)



import re

class Entity:
	def __init__(self, start, end, entity_id=None, quote_id=None, quote_eid=None, proper=None, ner_cat=None, in_quote=None, text=None):
		# print("creating")
		self.start=start
		self.end=end
		self.entity_id=entity_id
		self.quote_id=quote_id
		self.proper=proper
		self.ner_cat=ner_cat
		self.in_quote=in_quote
		self.quote_eid=quote_eid
		self.text=text
		self.quote_mention=None
		self.global_start=None
		self.global_end=None

	def __str__(self):
		return ("%s %s %s %s %s %s %s %s" % (self.global_start, self.global_end, self.entity_id, self.proper, self.ner_cat, self.in_quote, self.quote_eid, self.text))


class Token:

	def __init__(self, paragraph_id, sentence_id, index_within_sentence_idx, token_id, text, pos, fine_pos, lemma, deprel, dephead, ner, startByte):
		self.text=text
		self.paragraph_id=paragraph_id
		self.sentence_id=sentence_id
		self.index_within_sentence_idx=index_within_sentence_idx
		self.token_id=token_id
		self.lemma=lemma
		self.pos=pos
		self.fine_pos=fine_pos
		self.deprel=deprel
		self.dephead=dephead
		self.ner=ner
		self.startByte=startByte
		self.endByte=startByte+len(text)
		self.inQuote=False
		self.event="O"

	def __str__(self):
		return '\t'.join([str(x) for x in [self.paragraph_id, self.sentence_id, self.index_within_sentence_idx, self.token_id, self.text, self.lemma, self.startByte, self.endByte, self.pos, self.fine_pos, self.deprel, self.dephead, self.event]])

	@classmethod 
	def convert(self, sents):
		toks=[]
		i=0
		cur=0
		for sidx, sent in enumerate(sents):
			for widx, word in enumerate(sent):
				token=Token(0, sidx,widx,i,word, None, None, None, None, None, None, cur)
				toks.append(token)
				i+=1
				cur+=len(word) + 1
		return toks

	@classmethod 
	def deconvert(self, toks):
		sents=[]
		sent=[]
		lastSid=None
		for tok in toks:
			if lastSid is not None and tok.sentence_id != lastSid:
				sents.append(sent)
				sent=[]
			sent.append(tok)
			lastSid=tok.sentence_id

		if len(sent) > 0:
			sents.append(sent)

		# print(sents)
		return sents


from spacy.tokens import Doc

class SpacyPipeline:
	def __init__(self, spacy_nlp):
		self.spacy_nlp=spacy_nlp
		self.spacy_nlp.max_length = 10000000


	def filter_ws(self, text):
		text=re.sub(" ", "S", text)
		text=re.sub("[\n\r]", "N", text)
		text=re.sub("\t", "T", text)
		return text
		
		
	def tag_pretokenized(self, toks, sents, spaces):

		doc = Doc(self.spacy_nlp.vocab, words=toks, spaces=spaces)
		for idx, token in enumerate(doc):
			token.sent_start=sents[idx]

		for name, proc in self.spacy_nlp.pipeline:
			doc = proc(doc)

		return self.process_doc(doc)

	def tag(self, text):

		doc = self.spacy_nlp(text)
		return self.process_doc(doc)

	def process_doc(self, doc):

		tokens=[]
		skipped_global=0
		paragraph_id=0
		current_whitespace=""
		sentence_id=0
		for sid, sent in enumerate(doc.sents):
			skipped_in_sentence=0
			skips_in_sentence=[]
			curSkips=0
			for w_idx, tok in enumerate(sent):
				if tok.is_space:
					curSkips+=1
				skips_in_sentence.append(curSkips)
	
			hasWord=False

			for w_idx, tok in enumerate(sent):

				if tok.is_space:
					skipped_global+=1
					skipped_in_sentence+=1
					current_whitespace+=tok.text
				else:
					if re.search("\n\n", current_whitespace) is not None:
						paragraph_id+=1

					hasWord=True

					head_in_sentence=tok.head.i-sent.start
					skips_between_token_and_head=skips_in_sentence[head_in_sentence]-skips_in_sentence[w_idx]
					token=Token(paragraph_id, sentence_id, w_idx-skipped_in_sentence, tok.i-skipped_global, self.filter_ws(tok.text), tok.pos_, tok.tag_, tok.lemma_, tok.dep_, tok.head.i-skipped_global-skips_between_token_and_head, None, tok.idx)
					tokens.append(token)
					current_whitespace=""

			if hasWord:
				sentence_id+=1

		return tokens

class StanzaPipeline:
	def __init__(self, nlp):
		self.nlp=nlp


	def filter_ws(self, text):
		text=re.sub(" ", "S", text)
		text=re.sub("[\n\r]", "N", text)
		text=re.sub("\t", "T", text)
		return text
		
		
	def tag(self, text):
#	def __init__(self, sentence_id, index_within_sentence_idx, token_id, text, pos, lemma, deprel, dephead, ner, startByte, spacyToken):

		text=re.sub("\s+", " ", text)
		doc = self.nlp(text)
		tokens=[]
		tid=0
		cur=0
		for sid, sent in enumerate(doc.sentences):
			for w_idx, tok in enumerate(sent.words):
				feats=tok.misc.split("|")
				start_char=-1
				for f in feats:
					parts=f.split("=")
					if parts[0] == "start_char":
						start_char=int(parts[1])

# paragraph_id, sentence_id, index_within_sentence_idx, token_id, text, pos, lemma, deprel, dephead, ner, startByte						
				# need to add paragraph ID in stanza here!!
				paragraph_id=-1
				token=Token(paragraph_id, sid, w_idx, tid, self.filter_ws(tok.text), tok.upos, tok.pos, tok.lemma, tok.deprel, cur+tok.head-1, None, start_char)
				tokens.append(token)
				tid+=1
			cur+=len(sent.words)

		return tokens





import torch, sys, re

from booknlp.english.bert_coref_quote_pronouns import BERTCorefTagger
import numpy as np
from booknlp.common.pipelines import Entity
from booknlp.english.name_coref import NameCoref
import pkg_resources

class LitBankCoref:

	def __init__(self, modelFile, gender_cats, pronominalCorefOnly=True):

		device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

		base_model=re.sub("google_bert", "google/bert", modelFile.split("/")[-1])
		base_model=re.sub(".model", "", base_model)

		self.model = BERTCorefTagger(gender_cats=gender_cats, freeze_bert=True, base_model=base_model, pronominalCorefOnly=pronominalCorefOnly)
		self.model.load_state_dict(torch.load(modelFile, map_location=device))
		self.model.to(device)
		self.model.eval()


	def tag(self, tokens, g_ents, refs, ref_gender, attributed_quotations, quotes):
		sentences, ents, max_words, max_ents=self.convert_data(tokens, g_ents)
		assignments,global_entities=self.test(sentences, ents, max_words, max_ents, refs, ref_gender, attributed_quotations, quotes)
		return assignments


	def test(self, test_doc, test_ents, max_words, max_ents, refs, ref_gender, attributed_quotations, quotes):

		global_entities=[]
		for ents in test_ents:
			global_entities.extend(ents)

		for ent in global_entities:
			if ent.in_quote:
				for idx, (q_start, q_end) in enumerate(quotes):
					if ent.global_start >= q_start and ent.global_start <= q_end:
						ent.quote_mention=attributed_quotations[idx]

		test_matrix, test_index, test_token_positions, test_ent_spans, test_starts, test_ends, test_widths, test_data, test_masks, test_transforms, test_quotes=self.model.get_data(test_doc, test_ents, max_ents, max_words)
		
		assignments=self.model.forward(test_matrix, test_index, existing=refs, token_positions=test_token_positions, starts=test_starts, ends=test_ends, widths=test_widths, input_ids=test_data, attention_mask=test_masks, transforms=test_transforms, ref_genders=ref_gender, entities=global_entities)
		
		aliasFile = pkg_resources.resource_filename(__name__, "data/aliases.txt")

		nameCoref=NameCoref(aliasFile)

		e_list=[]
		for ent in global_entities:
			e_list.append((ent.global_start, ent.global_end, "%s_%s" % (ent.proper, ent.ner_cat), ent.text))

		assignments=nameCoref.cluster_noms(e_list, assignments)

		for ass in assignments:
			if ass == -1:
				print(assignments)
				sys.exit(1)

		return assignments, global_entities

	def convert_data(self, tokens, entities):

		max_words=0
		max_ents=0

		sents=[]
		o_sents=[]
		sent=[]
		o_sent=[]
		lastSid=None
		max_sentence_length=500
		length=0
		mapper={}

		for tok in tokens:

			wptok=tok.text
			if wptok[0].lower() != wptok[0]:
				wptok="[CAP] " + wptok.lower()

			toks=self.model.tokenizer.tokenize(wptok)
			if lastSid is not None and (tok.sentence_id != lastSid or length + len(toks) > max_sentence_length):
				sents.append(sent)
				o_sents.append(o_sent)
				sent=[]
				o_sent=[]
				length=0
			
			sent.append(toks)
			o_sent.append(tok)

			lastSid=tok.sentence_id
			length+=len(toks)
		
		sents.append(sent)
		o_sents.append(o_sent)

		sentences=[]
		o_sentences=[]

		max_sentence_length=500

		sentence=[ "[CLS]" ]
		ents=[]
		o_sent=[]

		running_length=0

		for idx, sent in enumerate(sents):

			length=0
			for toks in sent:
				length+=len(toks)

			if length + running_length >= max_sentence_length:
				sentence.append("[SEP]")
				sentences.append(sentence)
				ents.append([])
				o_sentences.append(o_sent)
				running_length=1
				sentence=["[CLS]"]
				o_sent=[]

			running_length+=length

			for word in o_sents[idx]:
				mapper[word.token_id]=len(sentences), len(sentence)
				
				wptok=word.text
				if wptok[0].lower() != wptok[0]:
					wptok="[CAP] " + wptok.lower()

				sentence.append(wptok)

			o_sent.extend(o_sents[idx])


		if len(sentence) > 1:		
			sentence.append("[SEP]")
			ents.append([])
			o_sentences.append(o_sent)
			sentences.append(sentence)

		sents=o_sentences

		lastS=-1
		
		entities=sorted(entities)

		for (start, end, cat, text) in entities:
			sent_id, w_in_sent_id_start=mapper[tokens[start].token_id]
			e_sent_id, w_in_sent_id_end=mapper[tokens[end].token_id]

			# if sent_id != e_sent_id:
			# 	print(sent_id,e_sent_id, "crossing sentence boundaries!" )
			# if sent_id < lastS:
			# 	print(sent_id, lastS, "non-monotonic!")

			lastS=sent_id

			inQuote=0
			if tokens[start].inQuote or tokens[end].inQuote:
				inQuote=1
			

			ent=Entity(w_in_sent_id_start, w_in_sent_id_end, in_quote=inQuote, quote_eid=None, entity_id=None, text=text)
			ner_parts=cat.split("_")
			ent.ner_cat=ner_parts[1]
			ent.proper=ner_parts[0]
			ent.global_start=start
			ent.global_end=end

			ents[sent_id].append(ent)



		max_words=max(len(sentence) for sentence in sentences)
		max_ents=max(len(ent) for ent in ents)

		return sentences, ents, max_words, max_ents


